import {world, system, Block, BlockPermutation, EntityEquippableComponent, EquipmentSlot, EntityTameableComponent} from "@minecraft/server"

world.beforeEvents.worldInitialize.subscribe(({blockComponentRegistry})=>{
    blockComponentRegistry.registerCustomComponent('mobfarm:wait_for_power', {
        onTick: ({block, dimension}) =>{
            const facing_direction = block.permutation.getState("minecraft:facing_direction")
            const hasPower = checkForPower(facing_direction, block)
            if(hasPower){
                block.setPermutation(BlockPermutation.resolve("mobfarm:fan_enabled", {'minecraft:facing_direction':facing_direction}))
            }
            if(!hasPower){
                block.setPermutation(BlockPermutation.resolve("mobfarm:fan_disabled", {'minecraft:facing_direction':facing_direction}))
            }
        }
    }),
    blockComponentRegistry.registerCustomComponent('mobfarm:fan_powered', {
        onTick: ({block, dimension}) =>{
            const facing_direction = block.permutation.getState("minecraft:facing_direction")
            let location = {...block.location}
            let direction = {x:0,y:0,z:0}     
            let horizonStrength = 1
            let verticalStrength = 0.1  
            switch(facing_direction){
                case "east":
                    location = {...block.west(1)}
                    location.z += 0.5
                    direction = {x:-1,y:0,z:0}
                    break
                case "west":
                    location = {...block.east(1)}
                    location.z += 0.5
                    direction = {x:1,y:0,z:0}             
                    break;
                case "north":
                    location = {...block.south(1)}
                    location.x += 0.5
                    direction = {x:0,y:0,z:1}                 
                    break;
                case "south":
                    location = {...block.north(1)}
                    location.z += 0.8
                    location.x += 0.5
                    direction = {x:0,y:0,z:-1}
                    break;
                case "up":
                    location = {...block.below(1)}
                    location.z += 0.5
                    location.x += 0.5
                    location.y -= 1.5
                    direction = {x:0,y:1,z:0}     
                    verticalStrength = -0.5           
                    break;
                case "down":
                    location = {...block.above(1)}
                    location.z += 0.5
                    location.x += 0.5
                    direction = {x:0,y:1,z:0}
                    verticalStrength = 0.5
                    break;
            }
            location.y += 0.5
            //world.sendMessage(`${facing_direction} - ${JSON.stringify(location)} - ${JSON.stringify(direction)}`)
            try{
                const getEntitiesFromRay = dimension.getEntitiesFromRay(location, direction, {maxDistance:8})
                for(let raycastHit of getEntitiesFromRay){
                    const entity = raycastHit.entity
                    entity.applyKnockback(direction.x, direction.z, horizonStrength, verticalStrength)
                }
            }catch(e){}
            
        }
    }),
    blockComponentRegistry.registerCustomComponent('mobfarm:grinder_toggle', {
        onPlayerInteract: ({block, player, dimension}) => {
            const itemHold = player.getComponent(EntityEquippableComponent.componentId).getEquipment(EquipmentSlot.Mainhand)
            if(itemHold){
                switch(itemHold.typeId){
                    case "minecraft:lever":
                        var grinderEntities = dimension.getEntitiesAtBlockLocation(block.location)
                        for(let grinderEntity of grinderEntities){                       
                            if(grinderEntity.typeId == "mobfarm:grinder_damage"){
                                if(!grinderEntity.hasComponent(EntityTameableComponent.componentId)) return
                                const tameComponent = grinderEntity.getComponent(EntityTameableComponent.componentId)
                                tameComponent.tame(player)
                                if(itemHold.amount == 1){
                                    player.getComponent(EntityEquippableComponent.componentId).setEquipment(EquipmentSlot.Mainhand, undefined)
                                }else{
                                    itemHold.amount -= 1
                                    player.getComponent(EntityEquippableComponent.componentId).setEquipment(EquipmentSlot.Mainhand, itemHold)
                                }
                            }
                        }
                    break

                    case "mobfarm:looting_chip":
                        var grinderEntities = dimension.getEntitiesAtBlockLocation(block.location)
                        for(let grinderEntity of grinderEntities){       
                            if(grinderEntity.typeId == "mobfarm:grinder_damage" && !grinderEntity.hasTag("upgraded")){
                                grinderEntity.triggerEvent("looting_upgrade")
                                if(itemHold.amount == 1){
                                    player.getComponent(EntityEquippableComponent.componentId).setEquipment(EquipmentSlot.Mainhand, undefined)
                                }else{
                                    itemHold.amount -= 1
                                    player.getComponent(EntityEquippableComponent.componentId).setEquipment(EquipmentSlot.Mainhand, itemHold)
                                }
                            }                            
                        }
                    break

                    case "mobfarm:power_chip":
                        var grinderEntities = dimension.getEntitiesAtBlockLocation(block.location)
                        for(let grinderEntity of grinderEntities){  
                            if(grinderEntity.typeId == "mobfarm:grinder_damage" && !grinderEntity.hasTag("upgraded")){
                                grinderEntity.triggerEvent("power_upgrade")
                                if(itemHold.amount == 1){
                                    player.getComponent(EntityEquippableComponent.componentId).setEquipment(EquipmentSlot.Mainhand, undefined)
                                }else{
                                    itemHold.amount -= 1
                                    player.getComponent(EntityEquippableComponent.componentId).setEquipment(EquipmentSlot.Mainhand, itemHold)
                                }
                            }               
                        }
                    break
                }   
            }            
        }
    }),
    blockComponentRegistry.registerCustomComponent('mobfarm:grinder_placed', {
        onPlace: ({block,dimension,previousBlock}) => {
            const blockType = block.typeId
            const location = {...block.location}  
            location.x += 0.5
            location.z += 0.5         
            const grinderEntity = dimension.spawnEntity("mobfarm:grinder_damage", location)
            switch(blockType){
                case "mobfarm:place_diamond_grinder":
                    grinderEntity.addTag("diamond")
                    break
                case "mobfarm:place_iron_grinder":
                    grinderEntity.addTag("iron")
                    break
                case "mobfarm:place_ender_grinder":
                    grinderEntity.addTag("ender")
                    break
            }
        }
    }),
    blockComponentRegistry.registerCustomComponent('mobfarm:grinder_powered', {
        onTick: ({block}) => {
            const rotateState = block.permutation.getState("mobfarm:rotate")
            if(rotateState == 0){
                block.setPermutation(block.permutation.withState("mobfarm:rotate", 1))
            }else{
                block.setPermutation(block.permutation.withState("mobfarm:rotate", 0))
            }
        }
    }),
    blockComponentRegistry.registerCustomComponent('mobfarm:suction_small', {
        onPlace: ({block, dimension}) => {
            const location = {...block}
            location.x += 0.5
            location.y += 0.5
            location.z += 0.5
            dimension.spawnEntity("mobfarm:suction_hopper", location)
            const suction_inventory = dimension.spawnEntity("mobfarm:suction_hopper_container", location)
            suction_inventory.nameTag = "Inventory"
        },
        onPlayerInteract: ({player, block}) => {
            const location = block.location
            player.runCommand(`dialogue open @e[r=1, c=1, x=${location.x}, y=${location.y}, z=${location.z}, type=mobfarm:suction_hopper] @s suction_hopper`)
        }
    }),
    blockComponentRegistry.registerCustomComponent('mobfarm:suction_big', {
        onPlace: ({block, dimension}) => {
            const location = {...block}
            location.x += 0.5
            location.y += 0.5
            location.z += 0.5
            dimension.spawnEntity("mobfarm:suction_hopper", location)
            const suction_inventory = dimension.spawnEntity("mobfarm:suction_hopper_container_large", location)
            suction_inventory.nameTag = "Inventory"
        },
        onPlayerInteract: ({player, block}) => {
            const location = block.location
            player.runCommand(`dialogue open @e[r=1, c=1, x=${location.x}, y=${location.y}, z=${location.z}, type=mobfarm:suction_hopper] @s suction_hopper`)
        }
    })
})

/**
 * 
 * @param {String} facing_direction 
 * @param {Block} block
 * @returns {boolean}
 */
function checkForPower(facing_direction, block){
    let hasPower = false
    let location = block.location
    const validBlocksAndStates = {
        lever:{
            name: "open_bit",
            type: "boolean" 
        },
        redstone_wire:{
            name: "redstone_signal",
            type: "number"
        },
        daylight_detector:{
            name: "redstone_signal",
            type: "number"
        },
        powered_repeater:{
            name: "no_state",
            type: "none"
        },
        comparator:{
            name: "output_lit_bit",
            type: "boolean"
        },
        redstone_block:{
            name: "no_state",
            type: "none"
        },
        redstone_torch:{
            name: "no_state",
            type: "none"
        },
        button:{
            name: "button_pressed_bit",
            type: "boolean"
        },
        detector:{
            name: "powered_bit",
            type: "boolean"
        }
    }
        
    switch(facing_direction){
        case "east":
            location = block.west(-1)
            break
        case "west":
            location = block.east(-1)
            break;
        case "north":
            location = block.south(-1)
            break;
        case "south":
            location = block.north(-1)
            break;
        case "up":
            location = block.below(-1)
            break;
        case "down":
            location = block.above(-1)
            break;
    }

    const poweredBlock = block.dimension.getBlock(location)
    const poweredBlockType = poweredBlock.typeId
    const poweredBlockStates = poweredBlock.permutation.getAllStates()

    for(let [blocksAndStateKey, blocksAndStateValue] of Object.entries(validBlocksAndStates)){
        if(poweredBlockType.includes(blocksAndStateKey)){
            const stateName = blocksAndStateValue.name
            const stateValue = blocksAndStateValue.type
            for(let [poweredBlockStateKey, poweredBlockStateValue] of Object.entries(poweredBlockStates)){
                if(stateName == poweredBlockStateKey){                 
                    switch(stateValue){
                        case "boolean":
                            if(poweredBlockStateValue == true) hasPower = true
                            break
                        case "number":
                            
                            if(poweredBlockStateValue >= 1) hasPower = true
                            break
                        case "none":
                            hasPower = true
                            break
                    }
                }
            } 
            break;
        }        
    }

    //world.sendMessage(`${JSON.stringify(location)} - ${facing_direction}`)

    return hasPower
}

system.afterEvents.scriptEventReceive.subscribe(({id, message, sourceEntity})=>{
    const command = [id, message]
    switch(command[0]){
        case "mf:state":
            switch(command[1]){
                case "down":
                    const location = {...sourceEntity.location}
                    location.y -= 1
                    var block = sourceEntity.dimension.getBlock(location)
                    var states = block.permutation.getAllStates()
                    for(let [key, value] of Object.entries(states)){
                        sourceEntity.sendMessage(`${key} = ${value}`)
                    }
                break;

                default:
                    var block = sourceEntity.dimension.getBlock(sourceEntity.location)
                    var states = block.permutation.getAllStates()
                    for(let [key, value] of Object.entries(states)){
                        sourceEntity.sendMessage(`${key} = ${value}`)
                    }
            }
        break;
    }
})